/*
 * LICENSED UNDER THE LGPL 2.1,
 * http://www.gnu.org/licenses/lgpl.html
 * Also, the author promises to never sue IBM for any use of this file.
 *
 */

package net.jakubholy.testing.dbunit.embeddeddb;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Level;
import java.util.logging.Logger;

import junit.framework.AssertionFailedError;

import org.dbunit.DBTestCase;
import org.dbunit.DatabaseTestCase;
import org.dbunit.IDatabaseTester;
import org.dbunit.PropertiesBasedJdbcDatabaseTester;
import org.dbunit.database.DatabaseConfig;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.Column;
import org.dbunit.dataset.CompositeDataSet;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.ITable;
import org.dbunit.dataset.ITableMetaData;
import org.dbunit.dataset.NoSuchTableException;
import org.dbunit.dataset.xml.XmlDataSet;
import org.dbunit.operation.DatabaseOperation;

/**
 * Parent class for DbUnit test cases that want to use an
 * embedded (Derby) database for testing instead of an external
 * standalone DB.
 * <p>
 * The class uses the embedded Derby DB. The necessary tables must be already created.
 * The database is expected to be in ./testData/testDB/ .
 * </p><p>
 * Override {@link #getDataSet()} if you want to load data from other file than the
 * default one.
 * </p><p>
 * Overrides the default DBUnit setting to use fully qualified table name (schema.table) - see
 * {@link #getDatabaseTester()}.
 * </p>
 *
 * <h3>Example of usage</h3>
 * <h4>Verify data in DB</h4>
 * <code>
 * ITable rowsToInsertTbl = getConnection().createQueryTable("rowToInsert", sqlGetRowToInsert);
 * assertEquals("There shall be exactly 1 row for the given id.", 1, rowsToInsertTbl.getRowCount());
 * String dbVal = (String) rowsToInsertTbl.getValue(0, "MY_EXAMPLE_COLUMN");
 * </code>
 *
 * @see DatabaseCreator#createDbSchemaFromDdl(java.sql.Connection)
 * @see DatabaseCreator#DDL_FILE_PATH
 * @see DatabaseCreator#main(String[])
 *
 * @author jholy
 *
 * See <a href="http://sourceforge.net/apps/mediawiki/jeeutils/index.php?title=DbUnit_Test_Skeleton">DbUnit_Test_Skeleton subproject of jeeutils at SourceForge</a>
 *
 */
public abstract class AbstractEmbeddedDbTestCase extends DBTestCase {

	/**
	 * File's Subversion info (version etc.).
	 * It's replacement upon commit must be enabled in svn properties.
	 */
	public static final String SVN_ID = "$Id: AbstractEmbeddedDbTestCase.java 47 2009-11-25 22:05:28Z malyvelky $";

	//############# TEST DATA CONFIG ################
    /** Name of the folder storing our test data set XML, DB initialization DDL and the derby db itself. */
	public static final String TEST_DATA_FOLDER = "testData";

	/**
	 * Name of our test data set XML file loaded by this class unless overridden by a subclass.
	 * The file should be stored in the folder {@link #TEST_DATA_FOLDER}.
	 */
    public static final String DBUNIT_TEST_DATA_SET_NAME = "dbunit-test_data_set.xml";
    //############# /TEST DATA CONFIG ################

	//############# JDBC CONNECTION INFO ################
	/**
	 * Path to the directory storing the test derby DB.
	 * May be absolute or relative to the test execution directory.
	 */
	public static final String DERBY_DB_PATH = TEST_DATA_FOLDER + File.separator + "testDB";

	/** Test DB connection info - url. */
	public static final String JDBC_URL = "jdbc:derby:" + DERBY_DB_PATH;

	/** Test DB connection info - driver class name. */
	public static final String JDBC_DRIVER = "org.apache.derby.jdbc.EmbeddedDriver";

	/** Test DB connection info - user name. */
	public static final String JDBC_USERNAME = "sa";

	/** Test DB connection info - password. */
	public static final String JDBC_PASSWORD = "";
	//############# /JDBC CONNECTION INFO ################

	private static final Logger LOG = Logger.getLogger(AbstractEmbeddedDbTestCase.class.getName());

    /** DBUnit internal stuff that takes care of connecting to the DB. */
    private IDatabaseTester tester;

    /**
     * Perform cleanup of tables before running any test.
     * <p>
     * All tables that we want to clear must be specified in the dataset XML file
     * and in the proper order not to violate integrity constraint, namely dependent
     * tables following the tables that they reference.
     */
    protected DatabaseOperation getSetUpOperation() throws Exception {
       return DatabaseOperation.CLEAN_INSERT;
    }

    protected DatabaseOperation getTearDownOperation() throws Exception {
       return DatabaseOperation.NONE;
    }

    /**
     * Initiate the database by clearing tables and filling them with data from the data set.
     * Only tables that are mentioned in the data set are cleared of data.
     * <p>
     * We've added some better error reporting in the case that the test DB hasn't been created yet.
     *
     * @see org.dbunit.DatabaseTestCase#setUp()
     */
    protected void setUp() throws Exception {
        try {
            super.setUp();
        } catch (NoSuchTableException e) {
            throw new Exception("No such table exception - have you created & initialized the embedded DB? The missing table is '" + e.getMessage() + "'", e);
        } catch (SQLException e) {
        	LOG.log(Level.WARNING, "If the initial DB cleanup failed because of nonexistant schema " +
        			"(usually same as logged-in user name, for derby SA), set the System property " +
        			"PropertiesBasedJdbcDatabaseTester.DBUNIT_SCHEMA to an existing schema. Error: " + e.getMessage());
        	throw e;
        }

    } /* setUp */

    /*
     * (non-Javadoc)
     * @see org.dbunit.DatabaseTestCase#getDataSet()
     * Override if you want to load data from another file or files
     * or if you want to use different format of the data sat (e.g. FlatXmlDataSet).
     */
    protected IDataSet getDataSet() throws Exception {
    	final String dataSetPath = getDataSetPath();
    	try {
	        return new CompositeDataSet(new IDataSet[]{
	                new XmlDataSet(new FileInputStream( dataSetPath ))
	        });
    	} catch (FileNotFoundException e) {
    		throw new Exception("The data set file '" + dataSetPath +
    				"' not found (if the path is relative, are you running the test fro mthe correct directory?)"
    				, e);
    	}
    } /* getDataSet */

    /**
     * Constructs path to the data set XML used to fill the DB with data prior to running tests.
     * <p>
     * Implementation note: As of now, the path is relative to the test execution directory, which shall
     * normally be the project's directory.
     *
     * @see #TEST_DATA_FOLDER
     * @see #DBUNIT_TEST_DATA_SET_NAME
     */
	protected String getDataSetPath() {
		return TEST_DATA_FOLDER + File.separator + DBUNIT_TEST_DATA_SET_NAME;
	}

    /*
     * (non-Javadoc)
     * Override to support fully qualified table names of the form {schema}.{table}.
     * @see org.dbunit.DatabaseTestCase#getDatabaseTester()
     */
    protected final IDatabaseTester getDatabaseTester() throws Exception {
        if (tester == null) {
            tester = createAndInitDatabaseTester();
        }
        return tester;
    } /* getDatabaseTester */

    /**
     * Create a database tester that supports fully qualified table names.
     * <p>
     * We also set the system properties that define the DB connection based on
     * the JDBC_* constants like {@link #JDBC_URL}.
     */
	final static IDatabaseTester createAndInitDatabaseTester() {

        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_DRIVER_CLASS, JDBC_DRIVER );
        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL, JDBC_URL );
        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_USERNAME, JDBC_USERNAME );
        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_PASSWORD, JDBC_PASSWORD );
        //System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_SCHEMA, "MY_DEFAULT_SCHEMA" );

		return new PropertiesBasedJdbcDatabaseTester() {
		    public IDatabaseConnection getConnection() throws Exception {
		        IDatabaseConnection conn = super.getConnection();
		        conn.getConfig().setFeature(DatabaseConfig.FEATURE_QUALIFIED_TABLE_NAMES, true);
		        return conn;
		    }
		};
	} /* createDatabaseTester */

	/**
     * Replace the content of the database from the provided data set XML file.
     * This is useful e.g. if, for a particular test method, you need some special
     * data in the database and you don't want to create a new test class for that.
     * This method does the same database re-initialization that happens in {@link DatabaseTestCase#setUp}.
     * <p>
     * Consider this to be an experimental method that may have yet undiscovered
     * side effects on DbUnit's state and behavior and use at your own risk.
     *
     * <h4>Usage example</h4>
     * <pre><code>
     * public void testFetchData_emtpyTable() throws Exception {
     * 		this.replaceDatabase("TestData/empty_data_set.xml");
     * 		// Call the tested method
     * 		final List data = this.target.fetchData();
     * 		// Verify results
     * 		assertEquals("empty data set =&gt; shall be empty", 0, data.size());
     * }
     *
     * // other test method may use the default data set w/o calling replaceDatabase
     * </code></pre>
     *
     * @param dataSetFile (required) File name of a DbUnit Data Set XML
     * @throws FileNotFoundException If the provided string isn't an existing file's name
     * @throws Exception thrown by {@link IDatabaseTester#onSetup()}
     */
    protected void replaceDatabase(final String dataSetFile) throws FileNotFoundException, Exception {
    	// The data to load into the db replacing the old ones
    	final IDataSet newDataSet = new CompositeDataSet(new IDataSet[]{
                 new XmlDataSet(new FileInputStream(dataSetFile))
         });
    	replaceDatabase( newDataSet );
    } /* replaceDatabase(fileName) */

    /**
     * See {@link #replaceDatabase(String)} for detailed description, here
     * we only describe the differences.
     * <p>
     * This variant replaces the data by the provided Data Set - which
     * may be constructed programmatically - instead of
     * loading them from a .xml file.
     *
     * <h4>Example</h4>
     * <pre><code>
     * import org.dbunit.dataset.*;
     * import org.dbunit.dataset.datatype.DataType;
     *
     * final DefaultTable table = new DefaultTable("my_test_schema.my_test_table", new Column[]{
	 *			new Column("id", DataType.INTEGER)
	 *			, new Column("some_text", DataType.VARCHAR) });
	 * table.addRow( new Object[] {new Integer(1), "hello!"} );
	 *
	 * final DefaultDataSet dataSet = new DefaultDataSet(table);
	 * this.replaceDatabase( dataSet );
     * </code></pre>
     *
     * @param newDataSet (required) the data set used to replace the data with
     * @throws Exception thrown by {@link IDatabaseTester#onSetup()}
     */
    protected void replaceDatabase(final IDataSet newDataSet) throws Exception {
    	// Clean and load the database
        final IDatabaseTester databaseTester = getDatabaseTester();
        databaseTester.setSetUpOperation( DatabaseOperation.CLEAN_INSERT );
        databaseTester.setDataSet( newDataSet );
        databaseTester.onSetup();
    } /* replaceDatabase(IDataSet) */

    /**
     * Remove all data from the given test table.
     * This is useful e.g. if you want to verify that your code behaves correctly
     * when there are no data.
     * @throws SQLException
     */
    protected void clearTable(final String tableName) throws SQLException {
    	final String sql = "delete from " + tableName;
    	try {
    		final Statement stmt = getConnection().getConnection().createStatement();
    		final int deleteCnt = stmt.executeUpdate(sql);
    		LOG.fine("clearTable(" + tableName + "): deleted rows: " + deleteCnt);
    	} catch (Exception e) {
    		LOG.log(Level.SEVERE, "clearTable: Failure to delete rows from the table '" + tableName + "'", e);
    		throw new SQLException("Failure to delete rows from the table '" +
    				tableName + "', reason:" + e);
    	}
    } /* clearTable */

    /**
	 * Compare actual and expected values of DB rows
	 * using the DbUnit's {@link ITable} interface to
	 * access the test database.
	 *
	 * <h4>Subclassing for convenience</h4>
	 * You can create a subclass to make verification of
	 * your data/objects more convenient, something like
	 * <pre><code>
	 * private class ProductRowComparator extends RowComparator {
	 *
	 * 	private final Product[] expected;
	 *
	 * 	public ProductRowComparator(final IDatabaseTester databaseTester, final Product[] expected) {
	 * 		this.expected = expected.clone();
	 * 		java.util.Arrays.sort(this.expected); // sort by productId
	 *		initResultTableAndColumns(databaseTester, "select productId, price from eshop.products order by productId");
	 * 	}
	 *
	 * 	public void assertRowsEqual()  throws AssertionFailedError, DataSetException {
	 * 		super.assertRowCount(expected.length);
	 * 		for (int i = 0; i < expected.length; i++) {
				compareNext(new Object[]{ expected[i].getProductId(), expected[i].getPrice() });
			}
	 * 	}
	 *
	 * }
	 * </code></pre>
	 * You would than use it as follows:
	 * <pre><code>
	 * final Product[] expected = new Product[] { ... };
	 * new ProductRowComparator(getDatabaseTester(), expected).assertRowsEqual();
	 * </code></pre>
	 */
	public static class RowComparator {

		private String[] columnNames;
		private ITable resultTable;
		private int currentRow = -1;

		/**
		 * A new comparator that creates the internal {@link ITable} from
		 * the provided SQL.
		 * @param databaseTester (required) necessary to read the test DB;
		 * 	see {@link DBTestCase#getDatabaseTester()}
		 * @param sql (required) the SQL to execute
		 * @throws SQLException
		 * @throws Exception
		 */
		public RowComparator(final IDatabaseTester databaseTester, final String sql) throws SQLException, Exception {
			initResultTableAndColumns(databaseTester, sql);
		} /* constructor(String) */

		/**
		 * Check expected values against the provided {@link ITable}.
		 * @param resultTable (required)
		 * @throws DataSetException
		 */
		public RowComparator(final ITable resultTable) throws DataSetException {
			initResultTableAndColumns(resultTable);
		} /* constructor(ITable) */

		/**
		 * You must call {@link #initResultTableAndColumns(ITable)} or
		 * {@link #initResultTableAndColumns(String)} if using this constructor.
		 */
		protected RowComparator() {}

		/**
		 * Use the SQL to create an ITable and call {@link #initResultTableAndColumns(ITable)}.
		 * @param databaseTester
		 * @param sql (required)
		 * @throws DataSetException
		 * @throws SQLException
		 * @throws Exception
		 */
		protected final void initResultTableAndColumns(final IDatabaseTester databaseTester, final String sql) throws DataSetException, SQLException, Exception {
			if (databaseTester == null) {
				throw new IllegalArgumentException("The argument databaseTester: IDatabaseTester may not be null.");
			}

			final ITable resultTable = databaseTester.getConnection().
				createQueryTable("rowComparatorTbl", sql);

			initResultTableAndColumns(resultTable);
		} /* initResultTableAndColumns(String) */

		/**
		 * Set {@link #resultTable} and use its metadata to initialize {@link #columnNames}.
		 * @param resultTable (required)
		 * @throws DataSetException
		 * @see {@link #extractColumnNames()}
		 */
		protected final void initResultTableAndColumns(final ITable resultTable) throws DataSetException {
			if (resultTable == null) {
				throw new IllegalArgumentException("ITable resultTable may not be null");
			}
			this.resultTable = resultTable;
			this.columnNames = extractColumnNames();
		} /* initResultTableAndColumns(ITable) */

		/**
		 * Extract column names from {@link #resultTable}.
		 * @throws DataSetException
		 * @see ITable#getTableMetaData()
		 * @see ITableMetaData#getColumns()
		 */
		private String[] extractColumnNames() throws DataSetException {
			final Column[] columns = resultTable.getTableMetaData().getColumns();
			String[] columnNames = new String[columns.length];

			for (int i = 0; i < columns.length; i++) {
				columnNames[i] = columns[i].getColumnName();
			}
			return columnNames;
		} /* extractColumnNames */

		/**
		 * @see #compareNext(String, Object[])
		 */
		public void compareNext(final Object[] expectedValues) throws AssertionFailedError, DataSetException {
			compareNext(null, expectedValues);
		} /* compareNext */

		/**
		 * Compare the next row (or the 1st one if first use) in this.{@link #resultTable} with
		 * the provided expected values.
		 * @param msg (optional) A message to add to the AssertionFailedErrors produced
		 * @param expectedValues (required) the values to be expected in the columns;
		 * the values must be in the same order as columns in the internal {@link ITable}.
		 *
		 * @see ITable#getTableMetaData()
		 * @see ITableMetaData#getColumns()
		 */
		public void compareNext(final String msg, final Object[] expectedValues) throws AssertionFailedError, DataSetException {

			final String errorMsg = (msg == null)? "" : msg;

			++currentRow;
			assertTrue("There is no next row, the row count is " + resultTable.getRowCount()
					, currentRow < resultTable.getRowCount());

			if (expectedValues == null) {
				throw new IllegalArgumentException("Object[] expectedValues may not be null");
			}
			if (columnNames.length != expectedValues.length) {
				throw new IllegalArgumentException("columnNames.length (" + columnNames.length +
						") shall be same as expectedValues.length (" + expectedValues.length + ")");
			}

			// Do the test
			for (int i = 0; i < columnNames.length; i++) {
				Object actual = resultTable.getValue(currentRow, columnNames[i]);
				assertEquals(errorMsg + " (row (starting from 0) " + currentRow + ", column '" + columnNames[i] + "')"
						, expectedValues[i], actual);
			}

		} /* compareNext */

		/**
		 * Assert that the number of rows in this.{@link #resultTable} is as expected.
		 */
		public void assertRowCount(final int expected) throws AssertionFailedError {
			assertEquals("There shall be " + expected + " rows in total. The SQL or test data is likely incorrect."
					, expected, resultTable.getRowCount());
		} /* assertRowCount */

		/** Return the internal result table. */
		public ITable getResultTable() {
			return resultTable;
		} /* getResultTable */

	} /* class RowComparator */
}