/*
 * LICENSED UNDER THE LGPL 2.1,
 * http://www.gnu.org/licenses/lgpl.html
 * Also, the author promises to never sue IBM for any use of this file.
 * 
 */

package net.jakubholy.testing.dbunit.embeddeddb;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Logger;

import org.dbunit.IDatabaseTester;
import org.dbunit.PropertiesBasedJdbcDatabaseTester;
import org.dbunit.database.IDatabaseConnection;

/**
 * Utility class for initializing the embedded Derby test database.
 * 
 * @see #createDbSchemaFromDdl(Connection)
 */
public class DatabaseCreator {
	
	/** Name of the DDL file. */
	public static final String DDL_FILE_NAME = "create_db_content.ddl";
	
	/** Path to the DDL file, normally below {@link AbstractEmbeddedDbTestCase#TEST_DATA_FOLDER}. */
	public static final String DDL_FILE_PATH = AbstractEmbeddedDbTestCase.TEST_DATA_FOLDER + File.separator + DDL_FILE_NAME;
	
	private static final Logger LOG = Logger.getLogger(DatabaseCreator.class.getName());
	
	/**
	 * Run a DDL on the (empty) test database to create the schemas and tables
	 * to fill with test date. 
	 * The DDL is read from the file {@link #DDL_FILE_PATH}.
	 * <p>
	 * WARNING: The code cannot handle all the stuff you may use in a DDL.
	 * More exactly everything that you can pass to {@link Statement#execute(String)}
	 * is OK, in addition to that we can handle empty lines and comment lines
	 * starting with '--'.
	 * 
	 * @param connection (required) A connection to the test db; e.g. {@link AbstractEmbeddedDbTestCase#getDatabaseTester()}#getConnection()#getConnection().
	 * @throws FileNotFoundException The DDL file not found
	 * @throws IOException Failed to read the DDL
	 * @throws SQLException Failure during execution of the DDL
	 * @throws Exception
	 */
    public static void createDbSchemaFromDdl(Connection connection) throws FileNotFoundException, IOException, Exception, SQLException {
    	if (connection == null) {
    		throw new IllegalArgumentException("The argument connection: java.sql.Connection may not be null.");
    	}
    	
        LOG.info("createDbSchemaFromDdl: Going to initialize the test DB by creating the schema there...");
        final String sql = readDdlFromFile();
        
        LOG.info("createDbSchemaFromDdl: DDL read: " + sql);
		executeDdl(connection, sql);
        
        LOG.info("createDbSchemaFromDdl: done");
    } /* createDbSchemaFromDdl */

    /**
     * Execute the ddlStatements on the connection.
     * @param connection (required) connection to the target database.
     * @param ddlStatements (required) DDL statements separated by semi-colon (';'). 
     * It shouldn't contain any -- comments as JDBC may not be able to ignore them as appropriate.
     * @throws SQLException
     */
	private static void executeDdl(Connection connection, final String ddlStatements)
			throws SQLException {
		java.sql.Statement ddlStmt = connection.createStatement();
        final String[] statements = ddlStatements.split(";");
        
        for (int i = 0; i < statements.length; i++) {
        	if (statements[i].trim().length() > 0) {
	            LOG.info("createDbSchemaFromDdl: Adding batch stmt: " + statements[i]);
	            ddlStmt.addBatch(statements[i]);
        	}
        }
        
        ddlStmt.executeBatch();
	} /* executeDdl */

    /**
     * Read the DDL from the file {@link #DDL_FILE_PATH} into a String.
     * @return the DDL
     * @throws FileNotFoundException
     * @throws IOException
     */
	private static String readDdlFromFile()
			throws FileNotFoundException, IOException {
		// DbUnit can't handle line comments within ddl => skip
        BufferedReader sqlReader = new BufferedReader(new FileReader(DDL_FILE_PATH));
        String line;
        StringBuffer sqlBuffer = new StringBuffer();
        while ((line = sqlReader.readLine()) != null) {
        	line = line.trim();
            if (line.length() > 0 && !line.startsWith("--")) {
                sqlBuffer.append(line).append("\n");
            }
        }
		return sqlBuffer.toString();
	} /* readDdlFromFile */

    /**
     * Run as a java console program to create and initialize the embedded Derby DB.
     * DB connection info is taken over from the {@link AbstractEmbeddedDbTestCase}.
     * <p>
     * To retry, simply delete the database (normally testData/testDB/).
     * 
     * @param args ignored
     * @throws Exception 
     * 
     * @see #createAndInitializeTestDb
     */
    public static void main(String[] args) throws Exception {
    	createAndInitializeTestDb();
	} /* main */

    /**
     * Create and from a DDL initialize the embedded Derby DB.
     * DB connection info is taken over from the {@link AbstractEmbeddedDbTestCase}.
     * <p>
     * To retry, simply delete the database (normally testData/testDB/).
     * 
     * @throws Exception 
     * 
     * @see #createDbSchemaFromDdl(java.sql.Connection)
     * @see #DDL_FILE_PATH
     * @see AbstractEmbeddedDbTestCase#createAndInitDatabaseTester()
     */
	private static void createAndInitializeTestDb() throws Exception,
			FileNotFoundException, IOException, SQLException {
		// Get a tester that we will use to access the DB;
    	// thus we're sure we access the same DB the tests will access
    	final IDatabaseTester databaseTester = AbstractEmbeddedDbTestCase.createAndInitDatabaseTester();
    	
    	// Modify the DB URL to tell Derby to create the DB if not existing
    	final String oldDbConnUrl = System.getProperty( 
    			PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL );
    	if (oldDbConnUrl == null) {
    		throw new IllegalStateException("The required property " + 
    				PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL +
    				" is not set (yet).");
    	}
    	// Create st. like "jdbc:derby:testData/testDB;create=true"
    	final String selfCreatingDbConnUrl = oldDbConnUrl + ";create=true"; // not very foolproof...
    	System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL, selfCreatingDbConnUrl );
    	
		// Get a connection
    	final IDatabaseConnection dbUnitConnection = 
    		databaseTester.getConnection();
    	
    	// Initialize the DB content
    	try {
    		DatabaseCreator.createDbSchemaFromDdl( dbUnitConnection.getConnection() );
    	} finally {
    		try {
				dbUnitConnection.close();
			} catch (SQLException e) {}
    	}
	} /* createAndInitializeTestDb */
	
}